#!/usr/bin/env python
"""
Utility to create a meaningful set of GDFS documents
from a number of ES indexes.
"""
import logging
import os
import sys
import time
import json
import subprocess
import operator

VERSION = '2.4.3'

# Set up logging.
log = logging.getLogger('{0}[{1}]'.format(os.path.basename(sys.argv[0]),
                                          os.getpid()))

blacklist_index = ['squirro_v7_fp', 'squirro_v7_filter']

def main():
    """
    The main method. Any exceptions are caught outside of this method and will
    be handled.
    """

    config = get_config()

    #get all config files
    es_data_path = config.get('elasticsearch_data_folder')

    if not es_data_path:
        log.error("Please specificy 'elasticsearch_data_folder' in there ini file.")
        sys.exit(1)

    target_folder = config.get('target_folder')
    if not target_folder:
        log.error("Please specificy 'target_folder' in there ini file.")
        sys.exit(2)

    shards = get_shard_folders(es_data_path)

    if not shards:
        log.error('No shards found in %r', es_data_path)
        sys.exit(1)

    indexes = config.get('indexes', 'all').lower()

    if indexes != 'all':
        indexes = indexes.split(' ')
        log.info("Using %s to create gfds files", indexes)
    else:
        indexes = []
        log.info("Using all indexes to create gfds files")


    #filter the files
    if len(indexes) == 0:
        usable_shards = shards
    else:
        usable_shards = []
        for shard in shards:
            for index in indexes:
                if shard.find(index) != -1:
                    usable_shards.append(shard)

    usable_shards.sort()
    log.info('Using these shard folders:')

    for shard in usable_shards:
        log.info('-> %s', shard)


    languages = config.get('languages', 'en de fr es it nl pt ru zh').strip().split(' ')
    frequency_lower_limit = int(config.get('frequency_lower_limit', 15))

    log.info('Using these languages:')

    for language in languages:

        log.info('-> %s', language)
        terms = {
            "number_of_documents": 0,
            "file":0,
            "total_files":1,
            "document_frequencies":{}
        }

        for shard in usable_shards:
            terms = extract_gdfs(shard, terms, language)

        #cleaning up

        remove_numbers = config.get('remove_numbers', 'True')

        if remove_numbers.lower() == 'false':
            remove_numbers = False

        filtered_terms = {}


        log.info('Found %i unfiltered terms', len(terms['document_frequencies']))
        freqency_fiter = 0
        numeric_filter = 0

        for term, count in terms['document_frequencies'].iteritems():

            #ignore low frequenzy terms to keep file small
            if count < frequency_lower_limit:
                freqency_fiter += 1
                continue

            if remove_numbers:
                #ignore numbers
                try:
                    float(term)
                    numeric_filter += 1
                    continue #we avoid pure numbers
                except ValueError:
                    pass

            filtered_terms[term] = count

        terms['document_frequencies'] = filtered_terms

        log.info("%i terms left after filtering", len(filtered_terms))
        log.info(" - %i got removed due to too low frequenzy", freqency_fiter)
        log.info(" - %i got removed due to being numbers", numeric_filter)

        segment_count = int(config.get('files_per_language', 8))

        log.info("Creating %i files per language", segment_count)

        segment_limit = int(len(terms['document_frequencies'])/segment_count)

        log.info("Up to %i terms per file", segment_limit)

        segments = []
        segment = {}

        term_count = 0
        for term, count in terms['document_frequencies'].iteritems():

            segment[term] = count
            term_count += 1

            if term_count == segment_limit:
                segments.append(segment)
                segment = {}
                term_count = 0

        target_path = config.get('target_folder', '/tmp')

        count = 0
        for segment in segments:

            terms['file'] = count
            terms['total_files'] = len(segments)
            terms['document_frequencies'] = segment

            file_path = '%s/%s%i.json' % (target_path, language, count)

            log.info('Writing %i terms into %s', len(terms['document_frequencies']), file_path)

            with open(file_path , 'w') as export_file:
                json.dump(terms, export_file, indent=4)

            count += 1


    log.info('All done!')

def extract_gdfs(index_path, terms, language):

    command = 'java -jar global_dfs.jar %s %s -o /tmp/%s > /dev/null' % (index_path, language, language)
    log.info('Executing: %s', command)
    code = subprocess.call(command, shell=True)

    if code != 0:
        log.warn('Skipping failed shard %s', index_path)
        return terms

    try:
        with open('/tmp/en0.json', 'r') as gdfs_file:
            gdfs_data = json.load(gdfs_file)

        log.info('Processed %i documents', gdfs_data['number_of_documents'])
        log.info('Found %i terms', len(gdfs_data['document_frequencies']))
    except IOError:
        log.warn('No terms found for %s in %r', language, index_path)
        return terms

    #count the documents
    terms['number_of_documents'] += gdfs_data['number_of_documents']

    for term, count in gdfs_data['document_frequencies'].iteritems():

        if terms['document_frequencies'].get(term):
            terms['document_frequencies'][term] += count
        else:
            terms['document_frequencies'][term] = count

    return terms


def get_shard_folders(es_data_path):
    """Collects all shard folders of all indexes"""

    log.info("Looking for shards in %r", es_data_path)
    log.info("Blacklisted: %r", blacklist_index)
    folders = []

    for root, subFolders, files in os.walk(es_data_path):
        for file in files:

            if file == 'write.lock':

                blacklisted = False
                for black_list in blacklist_index:

                    if root.find(black_list) != -1:
                        blacklisted = True #not interested in blacklisted indexes

                if not blacklisted:
                    folders.append(root)

    return folders


def get_config():
    """parse the ini file"""

    config = {}

    with open('create_gdfs.ini', 'r') as ini_file:
        for line in ini_file:
            if line[0] == '#':
                continue #no comments

            if line.find('=') == -1:
                continue #no param here

            parts = line.split('=')

            param_name = parts[0].strip()
            param_value = "=".join(parts[1:]).strip()

            config[param_name] = param_value

    return config

def setup_logging():
    """Set up logging based on the command line options.
    """
    # Set up logging
    fmt = '%(asctime)s %(name)s %(levelname)-8s %(message)s'
    level = logging.INFO
    logging.getLogger(
        'requests.packages.urllib3.connectionpool').setLevel(logging.WARN)

    logging.basicConfig(level=level, format=fmt)

    # Log time in UTC
    logging.Formatter.converter = time.gmtime


# This is run if this script is executed, rather than imported.
if __name__ == '__main__':
    setup_logging()

    log.info('Starting process (version %s).', VERSION)

    # run the application
    try:
        main()
    except Exception as e:
        log.exception('Processing error')
